/*
 *    Copyright  1996, 1997, 1998, 1999, 2000 by Remco Treffkorn
 *    Distributed under the terms of the Q Public License. Please see
 *    the included file LICENSE.
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <Xm/DialogS.h>
#include <Xm/PushBG.h>
#include <Xm/PushB.h>
#include <Xm/LabelG.h>
#include <Xm/PanedW.h>
#include <Xm/Form.h>
#include <Xm/RowColumn.h>
#include <Xm/TextF.h>
#include <Xm/ToggleB.h>
#include <Xm/Frame.h>
#include <Xm/Label.h>
#include <Xm/List.h>
#include <Xm/ScrolledW.h>
#include "Outline.h"
#include "Handle.h"
#include "xmap.h"
#include "debug.h"

extern int open_tiff_map(struct map *map);
extern int bsb_open_kap(char *fn);
extern void get_map(struct map *map);

extern Widget toplevel;
extern Widget status_1;
extern Pixmap dir_opened, dir_closed, mappmap;
extern void open_db(void);
extern struct db_entry *read_db(int);
extern void close_db(void);
extern void schedule_map_update(char *);

struct map map;

typedef struct {
    char *label;
    void (*callback) ();
    XtPointer data;
} ActionAreaItem;

void new_map(char *map);

static void close_dialog(), end_dialog(), ok_pushed(), help();

static Widget list_w, mapdialog, outline;
static Widget selected;
static struct db_entry *item;

extern Widget GetTopShell(Widget w);
extern Widget CreateActionArea(Widget w, ActionAreaItem * actions,
			       int num_act);

static int inhere = 0;

static String translations = "#override\n\
        <Key>Down:      ManagerGadgetTraverseRight()\n\
        <Key>Up:        ManagerGadgetTraverseLeft()\n\
        <Key>End:       ManagerGadgetTraverseHome() \
                        ManagerGadgetTraverseLeft()\n";


void import_map() 
{
}

void PushButtonHandleCB(Widget w, XtPointer closure, XtPointer call_data)
{
    XmPushButtonCallbackStruct *pb =
	(XmPushButtonCallbackStruct *) call_data;

    Widget label = (Widget) closure, parent, outline;

    if (pb->click_count == 2) {

	/* Lookup the parents outline widget. */
	parent = XtParent(w);

	XtVaGetValues(parent, XmNsubWidget, &outline, NULL);

	if ((outline != NULL) && XtIsManaged(outline)) {
	    XtUnmanageChild(outline);
	    XtVaSetValues(label, XmNlabelPixmap, dir_closed, NULL);
	} else if (outline != NULL) {
	    XtManageChild(outline);
	    XtVaSetValues(label, XmNlabelPixmap, dir_opened, NULL);
	}
    }
}



void mapCB(Widget w, XtPointer ud, XtPointer call_data)
{
    XmPushButtonCallbackStruct *pb =
	(XmPushButtonCallbackStruct *) call_data;
    Pixel foreground, background;

    if (pb->click_count == 2) {
	ok_pushed(w, ud, call_data);
    }
    XtVaGetValues(XtParent(w),
		  XmNforeground, &foreground,
		  XmNbackground, &background, NULL);

    if (item) {
	XtVaSetValues(selected,
		      XmNforeground, foreground,
		      XmNbackground, background, NULL);
    }
    selected = w;
    item = (struct db_entry *) ud;

    XtVaSetValues(w,
		  XmNforeground, background,
		  XmNbackground, foreground, NULL);
}

static void add_entry(Widget outline, struct db_entry *db)
{
    XmString s;
    Widget handle, label, button;

    s = XmStringCreateSimple(db->description);
    handle = XtVaCreateManagedWidget("f", xmHandleWidgetClass, outline,
				     XmNnavigationType, XmNONE,
				     XmNtranslations,
				     XtParseTranslationTable(translations),
				     XmNspacing, 0, NULL);

    label =
	XtVaCreateManagedWidget("label", xmPushButtonWidgetClass, handle,
				XmNlabelType, XmPIXMAP, XmNlabelPixmap,
				mappmap, XmNshadowThickness, 0,
				XmNmultiClick, XmMULTICLICK_KEEP,
				XmNspacing, 0, NULL);
    XtAddCallback(label, XmNactivateCallback, mapCB, (XtPointer) db);

    button =
	XtVaCreateManagedWidget("button", xmPushButtonWidgetClass, handle,
				XmNlabelString, s, XmNshadowThickness, 0,
				XmNmultiClick, XmMULTICLICK_KEEP,
				XmNspacing, 0, NULL);
    XtAddCallback(button, XmNactivateCallback, mapCB, (XtPointer) db);
    XmStringFree(s);
}

static Widget new_outline(Widget w, char *p, int flag)
{
    XmString s;
    Widget outline, handle, label, button;

    s = XmStringCreateSimple(p);
    outline = XtVaCreateManagedWidget("outline", xmOutlineWidgetClass, w,
				      XmNindentation, 0,
				      XmNoutline, FALSE,
				      XmNnavigationType, XmTAB_GROUP,
				      XmNspacing, 0,
				      XmNtranslations,
				      XtParseTranslationTable
				      (translations), NULL);

    handle = XtVaCreateManagedWidget("root", xmHandleWidgetClass, outline,
				     XmNnavigationType, XmNONE,
				     XmNtranslations,
				     XtParseTranslationTable(translations),
				     XmNspacing, 0, NULL);

    label =
	XtVaCreateManagedWidget("label", xmPushButtonWidgetClass, handle,
				XmNlabelType, XmPIXMAP, XmNlabelPixmap,
				flag ? dir_opened : dir_closed,
				XmNshadowThickness, 0, XmNmultiClick,
				XmMULTICLICK_KEEP, XmNspacing, 0, NULL);

    button =
	XtVaCreateManagedWidget("button", xmPushButtonWidgetClass, handle,
				XmNlabelString, s, XmNshadowThickness, 0,
				XmNmultiClick, XmMULTICLICK_KEEP, NULL);
    outline =
	XtVaCreateWidget("outline", xmOutlineWidgetClass, handle,
			 XmNindentation, 20, XmNoutline, TRUE,
			 XmNnavigationType, XmNONE, XmNspacing, 0,
			 XmNtranslations,
			 XtParseTranslationTable(translations), NULL);

    if (flag)
	XtManageChild(outline);

    XtVaSetValues(handle, XmNsubWidget, outline, NULL);
    XtAddCallback(label, XmNactivateCallback, PushButtonHandleCB,
		  (XtPointer) label);
    XtAddCallback(button, XmNactivateCallback, PushButtonHandleCB,
		  (XtPointer) label);
    XmStringFree(s);
    return outline;
}

void select_map_dialog(Widget w, XtPointer client_data,
		       XtPointer call_data)
{
    int mapno = 0;
    Position x, y;
    struct db_entry *p;
    Widget pane, rc, action_a;
    extern Widget CreateActionArea();
    static ActionAreaItem action_items[] = {
	{"OK", ok_pushed, NULL},
	{"Cancel", close_dialog, NULL},
	{"Help", help, "Help Button"},
    };

    if (inhere)
	return;
    inhere++;

    selected = (Widget) NULL;

    /* The DialogShell is the Shell for this dialog.  Set it up so
     * that the "Close" button in the window manager's system menu
     * destroys the shell (it only unmaps it by default).
     */
    mapdialog = XtVaCreatePopupShell("mapdialog",
				     xmDialogShellWidgetClass,
				     GetTopShell(w), XmNtitle,
				     "Select a Map",	/* give arbitrary title in wm */
				     XmNdeleteResponse, 
				     XmDESTROY,	/* system menu "Close" action */
				     XmNheight, 305,
				     XmNwidth, 330,
				     XmNminHeight, 305,
				     XmNminWidth, 330,
				     XmNmaxHeight, 305,
				     XmNmaxWidth, 330, NULL);
    XtAddCallback(mapdialog, XmNdestroyCallback, (void *) end_dialog,
		  NULL);

    /* Create the paned window as a child of the dialog.  This will
     * contain the control area and the action area
     * (created by CreateActionArea() using the action_items above).
     */
    pane = XtVaCreateWidget("pane", xmPanedWindowWidgetClass, mapdialog,
			    XmNsashWidth, 1, XmNsashHeight, 1, NULL);

    /* create the control area which contains a
     * Label gadget and a TextField widget.
     */
    rc = XtVaCreateWidget("mapbul", xmFormWidgetClass, pane,
			  XmNnoResize, True,
			  XmNshadowThickness, 0,
			  XmNheight, 305, XmNwidth, 330, NULL);


//////////////////////////////////////////////////////////////////////////////////
    list_w =
	XtVaCreateManagedWidget("scroller", xmScrolledWindowWidgetClass,
				rc, XmNscrollingPolicy, XmAUTOMATIC,
				XmNscrollBarDisplayPolicy, XmAS_NEEDED,
				XmNbottomAttachment, XmATTACH_FORM,
				XmNtopAttachment, XmATTACH_FORM,
				XmNleftAttachment, XmATTACH_FORM,
				XmNrightAttachment, XmATTACH_FORM, NULL);

//  XtAddCallback(scroller, XmNtraverseObscuredCallback, traverseCB, NULL);

    open_db();
//

    outline = new_outline(list_w, "Maps", True);

    item = NULL;

    while ((p = read_db(mapno++))) {
      nextbsb:
	if (p->flags & M_BSBDIR) {
	    Widget so;
	    so = new_outline(outline, p->description, False);
	    while ((p = read_db(mapno++)) && (p->flags & M_BSBFILE)
		   && (p->flags & M_BSBSHEET)) {
		add_entry(so, p);
	    }
	    if (!p)
		continue;
	    goto nextbsb;
	}
	add_entry(outline, p);
    }

    XtManageChild(list_w);

//////////////////////////////////////////////////////////////////////////////////

    XtVaGetValues(toplevel, XmNx, &x, XmNy, &y, NULL);

    XtVaSetValues(mapdialog, XmNx, x, XmNy, y, NULL);


    /* RowColumn is full -- now manage */
    XtManageChild(rc);

    /* Create the action area. */
    action_a =
	CreateActionArea(pane, action_items, XtNumber(action_items));

    XtManageChild(pane);

    XtManageChild(mapdialog);
    XtPopup(mapdialog, XtGrabNone);
}

static void end_dialog(Widget w, XtPointer client_data,
		       XtPointer call_data)
{
    close_db();
    inhere = 0;
}

static void close_dialog(Widget w, XtPointer client_data,
			 XtPointer call_data)
{
    XtDestroyWidget(mapdialog);
    end_dialog(w, mapdialog, call_data);
}


static int 
check_map(char **filename, int flag)
{
    int fd;
    struct stat sb;
    char bufp[256];

    if (flag & M_BSBFILE)
	strtolower(*filename);
    snprintf(bufp, 255, "%s/%s", get_mapdir(), *filename);

    if (stat(bufp, &sb) == -1) {
	snprintf(bufp, 255, "  File does not exist: \n\n%s/%s\n", get_mapdir(),
		*filename);
	error(bufp, 0);
	return 0;
    }
    if ((flag & M_SHAPE)) {
	if (!S_ISDIR(sb.st_mode)) {
	    snprintf(bufp, 255, "  Not a directory: \n\n%s/%s", get_mapdir(),
		    *filename);
	    error(bufp, 0);
	    return 0;
	}
	snprintf(bufp, 255, "%s/default.aep", *filename);
	free(*filename);
	*filename = strdup(bufp);
	snprintf(bufp, 255, "%s/%s", get_mapdir(), *filename);

    } else {
	if (!S_ISREG(sb.st_mode)) {
	    snprintf(bufp, 255, "  Not a regular file: \n\n%s/%s", 
		     get_mapdir(), *filename);
	    error(bufp, 0);
	    return 0;
	}
    }
    if ((fd = open(bufp, O_RDONLY) == -1)) {
	snprintf(bufp, 255, "  Can't open for read: \n\n%s/%s", get_mapdir(),
		*filename);
	error(bufp, 0);
	return 0;
    }

    close(fd);

    return 1;
}

static void ok_pushed(Widget w, XtPointer client_data, XtPointer call_data)
{
    char *filename;

    if (item == NULL) {
	beep();
	return;
    }
    if (!check_map(&item->name, item->flags)) {
	beep();
	return;
    }
    //fprintf(stderr, "item->name = %s\n", item->name);

    mapdata.calibrated = False;
    if (item->flags & M_BSBFILE)
	mapdata.type = BSB;
    else if (item->flags & M_SHAPE)
	mapdata.type = SHAPE;
    else
	mapdata.type = SRASTER;

    filename = strdup(item->name);

    XtDestroyWidget(mapdialog);

    schedule_map_update(filename);
    inhere = 0;
}

static void help(Widget w, XtPointer client_data, XtPointer call_data)
{
    do_help();
}

void get_line(FILE * fp, char *p)
{
    *p = '\0';
    fgets(p, 255, fp);
    while (*p && *p != '\n' && *p != '\r')
	p++;
    *p = '\0';
}

void eval_name(char *p)
{
    if (strlen(p) >= 4 && strcmp(&(p[strlen(p) - 4]), ".kap") == 0)
	mapdata.type = BSB;

    if (strlen(p) >= 4 && strcmp(&(p[strlen(p) - 4]), ".aep") == 0)
	mapdata.type = SHAPE;
//fprintf(stderr,"Eval: %s type=%d\n",p,mapdata.type);
}

static void (*clnfunc) ();

void register_cleanup(void *p)
{
    clnfunc = p;
}

static void cleanup()
{
    if (clnfunc)
	clnfunc();
    clnfunc = NULL;
}


//
// Verify that the mapdir environment is viable.
// Warn for minor problems, die on serious ones.
//

void verify_mapdirs(void)
{
    char bufp[256];
    struct stat fstat;

    if (stat(get_mapdir(), &fstat) == -1) {
	snprintf(bufp, 255, "There is a problem with your map "
		"directory\n%s\n\nAborting...\n", get_mapdir());
	error_wait(bufp, errno);
    }
    if (!S_ISDIR(fstat.st_mode)) {
	snprintf(bufp, 255, "%s is listed as your map directory,\n"
		"but it is not a directory!\n\nAborting...", get_mapdir());
	error_wait(bufp, 0);
    }
    snprintf(bufp, 255, "%s/map.db", get_mapdir());

    if (stat(bufp, &fstat) == -1) {
	snprintf(bufp, 255, "There is a problem with your map "
		"database\n%s/map.db\n\nAborting...\n", get_mapdir());
	error_wait(bufp, errno);
    }
    if (!S_ISREG(fstat.st_mode)) {
	snprintf(bufp, 255, "%s/map.db is listed as your map database,\n"
		"but it is not a regular file!\n\nAborting...", get_mapdir());
	error_wait(bufp, 0);
    }
}

void cond_free(void **what)
{
    if (*what) {
	free(*what);
	*what = NULL;
    }
}

//
// This is the first action taken when a new map is about to be loaded.
// Initialize the map structure.
//
//
void new_map(char *filename)
{
    FILE *fp;
    char *p;
    char lbuf[256];
    char bufp[256];
    int mflag = 0;

    if (!strcmp(filename, ""))
	return;

    eval_name(filename);
    switch (mapdata.type) {
	case BSB:
	    /* Strictly speaking we should find out if it's a file, 
	       but this should be safe -- djb */
	    mflag |= M_BSBFILE;
	    break;
	case SHAPE:
	    mflag |= M_SHAPE;
	    break;
    }
    if (check_map(&filename, mflag) == 0) {
	mapdata.type = 0;
	GetPrivateProfileString("Maps", "default", "world.ini", lbuf, 255,
				inifile);
	filename = lbuf;
    }
    snprintf(bufp, 255, "%s/%s", get_mapdir(), filename);

    cleanup();

    mapdata.zoom = 1;
    mapdata.mx_offset = 0;
    mapdata.my_offset = 0;
    mapdata.calibrated = 0;
    mapdata.calstate = 0;
    mapdata.map = &map;

    /* This assumes a linear edge model for maps */
    map.wedge = map.eedge = map.nedge = map.sedge = 0;
    cond_free((void **) &map.desname);
    cond_free((void **) &map.description);
    cond_free((void **) &map.filename);
    cond_free((void **) &map.ll);
    cond_free((void **) &map.ref1);
    cond_free((void **) &map.ref2);
    cond_free((void **) &map.waw);
    cond_free((void **) &map.nmap);
    cond_free((void **) &map.nemap);
    cond_free((void **) &map.emap);
    cond_free((void **) &map.semap);
    cond_free((void **) &map.smap);
    cond_free((void **) &map.swmap);
    cond_free((void **) &map.wmap);
    cond_free((void **) &map.nwmap);

    if (mapdata.type == BSB) {
	if (bsb_get_description(get_mapdir(), filename, &map) == TRUE) {
	    mapdata.zoom = 4;
	    ch_title(map.description);
	    get_map(&map);
	}
    } else if (mapdata.type == SHAPE) {
	//fprintf(stderr, "SHAPE desname = %s\n", filename);
	map.desname = strdup(filename);
	map.description = strdup("ShapeExplorer");
	ch_title(map.description);
	get_map(&map);
    } else {
	int i;
	char buf2[1024];

	map.desname = strdup(filename);
	i = strlen(map.desname);
	if (!strcmp(map.desname + (i - 4), ".ini")) {
	    snprintf(buf2, 1023, "%s/%s", get_mapdir(), map.desname);
	    GetPrivateProfileString("general", "maptitle", "", bufp, 128,
				    buf2);
	    map.description = strdup(bufp);
	    GetPrivateProfileString("general", "filename", "", bufp, 128,
				    buf2);
	    map.filename = strdup(bufp);
	    if (strlen(map.filename) == 0)
		mapdata.type = VECTOR;
	    else {
		GetPrivateProfileString("general", "format", "", bufp, 128,
					buf2);
		if (strcmp(bufp, "tiff") == 0)
		    mapdata.type = DRG;
	    }

	    GetPrivateProfileString("neighbors", "north", "", bufp, 128,
				    buf2);
	    map.nmap = strdup(bufp);
	    GetPrivateProfileString("neighbors", "east", "", bufp, 128,
				    buf2);
	    map.emap = strdup(bufp);
	    GetPrivateProfileString("neighbors", "south", "", bufp, 128,
				    buf2);
	    map.smap = strdup(bufp);
	    GetPrivateProfileString("neighbors", "west", "", bufp, 128,
				    buf2);
	    map.wmap = strdup(bufp);
	    GetPrivateProfileString("neighbors", "northeast", "", bufp, 128,
				    buf2);
	    map.nemap = strdup(bufp);
	    GetPrivateProfileString("neighbors", "southeast", "", bufp, 128,
				    buf2);
	    map.semap = strdup(bufp);
	    GetPrivateProfileString("neighbors", "southwest", "", bufp, 128,
				    buf2);
	    map.swmap = strdup(bufp);
	    GetPrivateProfileString("neighbors", "northwest", "", bufp, 128,
				    buf2);
	    map.nwmap = strdup(bufp);
	    map.nedge = GetPrivateProfileFloat("edges", "north", 0.0,
					       buf2);
	    map.eedge = GetPrivateProfileFloat("edges", "east", 0.0, buf2);
	    map.sedge = GetPrivateProfileFloat("edges", "south", 0.0,
					       buf2);
	    map.wedge = GetPrivateProfileFloat("edges", "west", 0.0, buf2);

	    GetPrivateProfileString("georef", "ref1", "0, 0, 0", bufp, 128,
				    buf2);
	    map.ref1 = strdup(bufp);

	    GetPrivateProfileString("georef", "ref2", "0, 0, 0", bufp, 128,
				    buf2);
	    map.ref2 = strdup(bufp);

	    GetPrivateProfileString("general", "waw", "*", bufp, 128, buf2);
	    map.waw = strdup(bufp);

	    /* Get a default to use, default to WGS84 */
	    mapdata.mapdatum = 
	      GetPrivateProfileInt("Maps", "datum", 100, inifile);

	    /* See if this map has override datum */
	    mapdata.mapdatum = 
	      GetPrivateProfileInt("general", "datum", mapdata.mapdatum, buf2);

	    ch_title(map.description);

	    if (mapdata.type == DRG) {
		/*
		 * Side effect: may reset datum based on GeoTIFF tags 
		 * if map is tagged NAD27 or NAD83
		 */
		if (open_tiff_map(&map)) {
		    mapdata.zoom = 4;
		} else {
		    mapdata.type = 0;
		    return;
		}
	    }

	    get_map(&map);
	} else if ((fp = fopen(bufp, "ro"))) {
	    if (strcmp(map.desname + (i - 4), ".des")) {
		/* I don't know what to do with this */
		error("Not a map description file!", 0);
		return;
	    }
	    map.desname = strdup(filename);

	    get_line(fp, bufp);	// Description

	    map.description = strdup(bufp);

	    get_line(fp, bufp);	// Map file

	    map.filename = strdup(bufp);
	    if (strlen(map.filename) == 0)
		mapdata.type = VECTOR;

	    get_line(fp, bufp);	// LL, 0, 0

	    map.ll = strdup(bufp);

	    get_line(fp, bufp);	// coord

	    map.ref1 = strdup(bufp);

	    get_line(fp, bufp);	// coord

	    map.ref2 = strdup(bufp);

	    get_line(fp, bufp);	// waw

	    map.waw = strdup(bufp);

	    fclose(fp);

	    ch_title(map.description);

	    if (!strcmp(map.ll, "DRG")) {
		if (open_tiff_map(&map)) {
		    mapdata.type = DRG;
		    mapdata.zoom = 4;
		} else {
		    mapdata.type = 0;
		    return;
		}
	    }
	    get_map(&map);
	} else {
	    error("No map description file\nfor this map!", 0);
	}
    }
}

geo_ref_tfw(FILE * fp, struct map *map)
{
    double n1, n2, e1, e2, sx, sy, t, lat, lon;
    int utmzone;
    char bufp[256];

    // x pixel scale
    get_line(fp, bufp);
    sscanf(bufp, "%lf", &sx);

    // translation better be 0!
    get_line(fp, bufp);
    sscanf(bufp, "%lf", &t);
    if (t != 0)
	return;

    // rotation better be 0!
    get_line(fp, bufp);
    sscanf(bufp, "%lf", &t);
    if (t != 0)
	return;

    // y pixel scale
    get_line(fp, bufp);
    sscanf(bufp, "%lf", &sy);

    // easting
    get_line(fp, bufp);
    sscanf(bufp, "%lf", &e1);

    // northing
    get_line(fp, bufp);
    sscanf(bufp, "%lf", &n1);

    mapdata.mx1 = 0;
    mapdata.my2 = 0;

    mapdata.mx2 = map->mwidth - 1;
    mapdata.my2 = map->mheight - 1;

    /* Should be map-specific .ini override for this, but why bother? */
    utmzone = GetPrivateProfileInt("General", "utmzone", 18, inifile);
    UTMtoDeg(utmzone, 0, e1, n1, &(mapdata.lat1), &(mapdata.lon1));
    UTMtoDeg(utmzone, 0, e1 + ((double) mapdata.mx2 * sx),
	     n1 + ((double) mapdata.my2 * sy),
	     &(mapdata.lat2), &(mapdata.lon2));

    XmTextFieldSetString(status_1, "World file geo-referenced map.");

    do_calibrate();
}
