/*
 *    Copyright  1996, 1997, 1998, 1999, 2000 by Remco Treffkorn
 *    Distributed under the terms of the Q Public License. Please see
 *    the included file LICENSE.
 */
/*****************************************************************************
 *  units.h - definition and conversion of map units                         *
 *****************************************************************************/
/* UNIT DEFINITIONS:
    SEC=Seconds of Arc (not used much here)
    DEG=(D)Degrees of Arc
     LU=(U)Lat/Long Units, the units used in the tiger files, which use these
        very small units to achieve 6 decimal point precision with integers.
    YPU=Latitude Projection Units, low-res latitude (rsh 16 bits) to space out 
        projection table entries
    XPU=Longitude Projection Coefficient Units, multiplier to project
        longitude to map X.  11 bit integer representation of 0<xp<1
     MU=Map units, intermediate cartesian coordinates of the projected
        pseudo-map.  Equal to LU for latitude, which is not vital
    PIX=(P)Pixels, coordinate units on the output map
   SPIX=Sub-Pixels, fractions of a pixel: acheives precision with integers
*/
/* Miscellaneous Constants and Macros */
#define PI      3.141592654
#define TWOPI   6.283185307
#define PIOVER2 1.5707963
#define DEG2RAD (3.1415926535897931160E0/180.0)
#define SEC_PER_DEG 3600        /* 3600 seconds per degree */
#define MI_PER_DEG 69.115
#define FT_PER_M 3.28
#define IN_PER_MU 4.379
#define AVGPPI 75  /* average pixels per inch (for 800x600, 14" screen) */
/* Real-world coordinates */
#define LU_PER_DEG  1000000     /* File lat/lon units per degree */
#define DEG_TO_LU(deg)      ( (long) ((deg) * LU_PER_DEG) )
#define LU_TO_DEG(lu)       ( (float) (lu) / LU_PER_DEG )
#define DEG90_LU    90000000 /* lu's in 90 degrees */
#define DEG180_LU   180000000/* lu's in 180 degrees */
#define DEG360_LU   360000000
/* Projection--RW coordinates to pseudo-map */
#define XPU_SHIFT   11  /* precision for X coefficients */
#define LU_YPU_SHIFT    9   /* precision reduction for latitude in projection table */
#define GEN_SHIFT       5       /* reduction of large longitudes to 21-bit */
#define LAT_TO_YPU(lu)  ( (lu) >> LU_YPU_SHIFT )
#define LAT_TO_XP(table,lat)    ( (table)[LAT_TO_YPU(lat)] )
#define LONU_TO_MU(lon,xp,shift) ( (((lon) >> (shift)) * (xp)) >> (XPU_SHIFT - (shift)) )
#define MU_TO_LONU(mu,xp,shift) (( ((mu) << (XPU_SHIFT - (shift))) / (xp) ) << (shift) ) 
/* Scaling--pseudo-map to map image */
#define SPIX_SHIFT  8       /* bits to shift PIX <-> SPIX */
#define MU_PER_SPIX(ms)     ((float) (ms)->muPerPix >> SPIX_SHIFT)
#define PIX_PER_DEG(ms)     ((float) LU_PER_DEG / (ms)->muPerPix)
#define MU_TO_SPIX(ms,mu)   (((mu) << SPIX_SHIFT) / (ms)->muPerPix)
#define MU_TO_PIX(ms,mu)    ((mu) / (ms)->muPerPix)
#define PIX_TO_MU(ms,pix)   ((pix) * (ms)->muPerPix)
#define SPIX_TO_MU(ms,spix) (((spix) * (ms)->muPerPix) >> SPIX_SHIFT)

